/* ipchains.c
 *
 * Minimal pseudo-cgi that allows manipulation of ipchains rules
 *  from a web browser.
 * 
 * Copyright (C) 2001 Michael McTernan,
 *   Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <netdb.h>
#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "ipfwc_kernel_headers.h" 
#include "libipfwc.h"
#include "ipchains-save.h"

/* Prototypes */
static void dump_rules(const char *chain);
static void print_policy(ip_chainlabel ipcl);
static void print_addrmask(struct in_addr adr,struct in_addr mask,int inv);
static void print_protocol(int proto,int inv);
static void print_interface(ip_chainlabel inf,int inv); 
static void print_ports(int p1,int p2,int inv); 
static void info(void);
static void print_tables(void);
static void delete_rule(void);
static void change_def(void);
static char *getoptionval(const char *option);
static char *getchainpolicy(const char *chain);
static void print_flags(int flags,int invflags);
static void print_flag(const char *flg,int cond,int inv);
static void makedirty();


/*main*/
int main(int argC,char **argV) {
  char *query;

  /* Get the query */
  query=getenv("QUERY_STRING");

  if(query!=NULL) {

    if(strstr(query,"action=")!=NULL) {

      if(strstr(query,"action=delete")!=NULL) 
        delete_rule();
      else if(strstr(query,"action=changedef")!=NULL) 
        change_def(); 

      return 0;
    }
  }

  printf("<h2>Firewall ruleset</h2>\n");
  info();
  print_tables();

  printf("<ul>"
         "<li><a href=\"/cgi-bin/ipchains-sane.cgi\">Apply sane ruleset</a></li>\n"
         "<li><a href=\"/ipchains-add.html\">Add rule</a><br>"
         "<br>"
         "<li><a href=\"/\">Main menu</a>"
         "</ul>");
 
  return 0;
 
}
 
/*getoptionval*/
char *getoptionval(const char *option) {
  char *query;
  char *s,*t;
 
  /* Get the query */
  query=getenv("QUERY_STRING");
  if(query==NULL)
    return NULL;
 
  /* Seek option */
  s=strstr(query,option);
  if(s==NULL)
    return NULL;
 
  /* Skip to end of option string */
  s+=strlen(option);
 
  /* Duplicate */
  s=strdup(s);
 
  /* Separate */
  t=strstr(s,"&");
  if(t!=NULL)
    *t='\0';
 
  return s;
 
}

void change_def() {
  char *chain;
  char *policy;
 
  printf("<h2>Default Policy</h2>");
 
  /* Get the chain */
  chain=getoptionval("chain=");
  if(chain==NULL) {
    printf("<p>Malformed request</p>");
  }
 
  /* Get the policy */
  policy=getoptionval("policy=");
 
  /* Prompt for policy if not given */
  if(chain!=NULL && policy==NULL) {
 
    printf("<form method=\"GET\">"
           "<p>The default policy for the <b>%s</b> chain should be\n"
           "<select name=\"policy\">\n"
           "<option value=\"ACCEPT\">Accept\n"
           "<option value=\"DENY\">Deny\n"
           "<option value=\"REJECT\">Reject\n"
           "</select>\n"
           "<input type=\"hidden\" name=\"action\" value=\"changedef\">\n"
           "<input type=\"hidden\" name=\"chain\" value=\"%s\">\n"
           "<input type=\"submit\" value=\"Ok\">\n"
           "</form></p>",chain,chain);
 
  }
 
  /* Set the policy if we have all info */
  if(chain!=NULL && policy!=NULL) {
 
    if(ipfwc_set_policy(chain,policy)==0) {
      printf("<p>An error occured setting the policy.</p>\n");
      perror(NULL);
    } else {
      printf("<p>The default policy for the %s chain has been set to %s</p>",
             chain,
             policy);
      ipchains_save();
    }
  }
 
  printf("<ul><li><a href=\"ipchains.c\">View ruleset</a><br>"
         "<br>\n"
         "<li><a href=\"/\">Main menu</a></li>\n"
         "</ul>");

  free(chain);
  free(policy);

  makedirty(); 
}
 
void delete_rule() {
  int            rule;
  char          *chain,*num;
 
  printf("<h2>Delete Firewall Rule</h2>\n");
 
  /* Get the chain */
  chain=getoptionval("chain=");
  num=getoptionval("rule=");
 
  if(chain==NULL || num==NULL || sscanf(num,"%d",&rule)!=1) {
    printf("<p>The HTTP request is malformed</p>\n");
  } else {
 
    if(ipfwc_delete_num_entry(chain,rule)==0)
      printf("<p>An error occured deleting the rule</p>\n");
    else {
      printf("<p>Rule deleted.</p>");
      ipchains_save();
    }
  }
 
  printf("<ul>"
         "<li><a href=\"ipchains.c\">Firewall Ruleset</a><br><br>\n"
         "<li><a href=\"/\">Main menu</a>\n"
         "</ul>");

  free(chain);
  free(num);

  makedirty(); 

}


/*info*/
static void info() {

  printf("<p>These rules determine how network data is handled when passing "
         "through the firwall, and can be used to prevent possibly malicious "
         "information from being accepted or routed. "
         "</p>"
         "<p>If you are unsure as to what rules should be set, you may "
         "select <a href=\"ipchains-sane.cgi\">this</a> for a general rule "
         "set that will generally serve most users' needs.</p>");
}


/*print_tables*/
static void print_tables() {

  dump_rules(IP_FW_LABEL_INPUT);
  dump_rules(IP_FW_LABEL_OUTPUT);
  dump_rules(IP_FW_LABEL_FORWARD);

}

/*dump_rules*/
/* Dump the rules in chain 
 */
static void dump_rules(const char *chain) {
  struct ipfwc_fwrule *rules;
  int                  nrules;
  int                  count=0;
  int                  t;

  /* Beautified heading */
  assert(strlen(chain)>1);
  printf("<h3>%c%s</h3>\n",toupper(*chain),&chain[1]);

  /* Get rules */
  rules=ipfwc_get_rules(&nrules,0);
  if(rules==NULL) {
    printf("<p>Error fetching rules.</p>");
    return;
  }

  /* Open table and write heading */
  printf("<table border=\"1\" cellpadding=\"2\" width=\"100%%\">"
         "<tr>"
         "<td><i><b>Policy</i></b></td>"
         "<td><i><b>Protocol</i></b></td>"
         "<td><i><b>Source</i></b></td>"
         "<td><i><b>Source Port</i></b></td>" 
         "<td><i><b>Destination</i></b></td>"
         "<td><i><b>Destination Port</i></b></td>" 
         "<td><i><b>Flags</i></b></td>" 
         "<td><i><b>Interface</i></b></td>"  
         "</tr>");

  /* Process rules */
  for(t=0;t<nrules;t++) {

    /* Check if rule is on required chain */
    if(strcmp(rules[t].chain->label,chain)==0) {
       
      /* Increase rule count for this chain */
      count++;
      
      /* Output row */
      printf("<tr>");

      printf("<td align=\"center\">"); 
        print_policy(rules[t].ipfw.label); 
      printf("</td>");

      printf("<td align=\"center\">");
        print_protocol(rules[t].ipfw.ipfw.fw_proto,rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_PROTO);
      printf("</td>");

      printf("<td align=\"center\">"); 
        print_addrmask(rules[t].ipfw.ipfw.fw_src,rules[t].ipfw.ipfw.fw_smsk,rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_SRCIP);
      printf("</td>");

      printf("<td align=\"center\">");
        print_ports(rules[t].ipfw.ipfw.fw_spts[0],rules[t].ipfw.ipfw.fw_spts[1],rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_SRCPT);
      printf("</td>");

      printf("<td align=\"center\">");
        print_addrmask(rules[t].ipfw.ipfw.fw_dst,rules[t].ipfw.ipfw.fw_dmsk,rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_DSTIP);
      printf("</td>"); 

      printf("<td align=\"center\">");
        print_ports(rules[t].ipfw.ipfw.fw_dpts[0],rules[t].ipfw.ipfw.fw_dpts[1],rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_DSTPT);
      printf("</td>");

      printf("<td align=\"center\">");
        print_flags(rules[t].ipfw.ipfw.fw_flg,rules[t].ipfw.ipfw.fw_invflg);
      printf("</td>");

      printf("<td align=\"center\">");
        print_interface(rules[t].ipfw.ipfw.fw_vianame,rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_VIA); 
      printf("</td>");

      /* Add remove rule button */
      printf("<td>"
             "<a href=\"ipchains.c?action=delete&chain=%s&rule=%d\">Remove</a>"
             "</td>",
             chain,count);

      printf("</tr>\n");

    }

  }

  /* Add the default policy */
  printf("<tr bgcolor=\"lightgrey\">");
  printf("<td colspan=\"8\">Default: <b>");
  print_policy(getchainpolicy(chain));
  printf("</b></td>");
 
  printf("<td>");
  printf("<a href=\"ipchains.c?action=changedef&chain=%s\">Change</a>",chain);
  printf("</td>\n");
 
  printf("</tr>\n");
 
  /* Close off table */
  printf("</table>");
}


/*getchainpolicy*/
static char *getchainpolicy(const char *chain) {
  struct ipfwc_fwchain *chains;
  int                   nchains;
  int                   t;

  /* Get all the chains */
  chains=ipfwc_get_chainnames(&nchains);

  for(t=0;t<nchains;t++) {
    
    if(strcmp(chains[t].label,chain)==0) 
      return strdup(chains[t].policy);
   
  }
  
  return NULL;
} 


/*print_protocol*/
/* Print the name of a protocol 
 */
static void print_protocol(int proto,int inv) {
  struct protoent *pe;

  /* Special case for 'any' */
  if(proto==0) {
    printf("any");
    return;
  }

  if(inv)
    printf("<i>not ");

  /* Try system approach */
  pe=getprotobynumber(proto); 
  if(pe!=NULL) {
    printf("%s",pe->p_name);
    // TODO: free(pe); ????
  } else {

    /* Hard wire common ones incase /etc/protocols is broken */
    switch(proto) {
    case IPPROTO_TCP:
      printf("TCP");
      break;
    case IPPROTO_UDP:
      printf("UDP");
      break;
    case IPPROTO_ICMP:
      printf("ICMP");
      break;
    default:
      printf("%u",proto);
    }

  }

  if(inv)
    printf("</i>");

}


/*print_addrmask*/
/* Print the address/mask in a pretty way
 */
static void print_addrmask(struct in_addr adr,struct in_addr mask,int inv) {

  if(inv)
    printf("<i>not ");

  if(mask.s_addr==0L) {
    printf("anywhere");
  } else {
    char *a=inet_ntoa(adr);

    printf("<a href=\"dnslookup.c?q=%s\">%s</a>",a,a);

    /* Add mask if valid */
    if(ntohl(mask.s_addr)!=0xFFFFFFFFL)
      printf("/%s",inet_ntoa(mask));
  }

  if(inv)
    printf("</i>");

}


/*print_policy*/
/* Print a ipchain policy or rule destination with colour 
 */
static void print_policy(ip_chainlabel ipcl) {

  if(ipcl==NULL) {
    printf("-");
    return;
  }

  if(strcmp("DENY",ipcl)==0)
    printf("<font color=\"red\">DENY</font>");
  else if(strcmp("REJECT",ipcl)==0)
    printf("<font color=\"red\">REJECT</font>");    
  else if(strcmp("ACCEPT",ipcl)==0)
    printf("<font color=\"green\">ACCEPT</font>");
  else if(strcmp("MASQ",ipcl)==0)
    printf("<font color=\"blue\">MASQUERADE</font>");
  else
    printf("%s",ipcl);

}


/*print_interface*/
/* Print an interface name.
 */
static void print_interface(ip_chainlabel inf,int inv) {

  if(inv)
    printf("<i>not ");

  if(*inf=='\0')
    printf("any");
  else if(strcmp(inf,"lo")==0)
    printf("loopback");
  else
    printf("%s",inf);

  if(inv)
    printf("</i>");

}


/*print_ports*/
/* Print the port range 
 */
static void print_ports(int p1,int p2,int inv) {

  if(inv)
    printf("<i>not ");

  if(p1==p2)
    printf("%d",p1);
  else if(p1==0 && p2==65535)
    printf("any");
  else
    printf("%d-%d",p1,p2);  

  if(inv)
    printf("</i>");

}

static void print_flags(int flags,int invflags) {

  print_flag("Log ",flags & IP_FW_F_PRN,0);
  print_flag("SYN ",flags & IP_FW_F_TCPSYN,invflags & IP_FW_INV_SYN);
  print_flag("Fragment ",flags & IP_FW_F_FRAG,invflags & IP_FW_INV_FRAG);

}

static void print_flag(const char *flg,int cond,int inv) {
  if(cond) {
    if(inv)
      printf("<I>not %s</I>",flg);
    else
      printf(flg);
  }
}


static void makedirty() {
  FILE *f;

  f=fopen("/etc/coyote/coyote.conf.dirty","w");
  if(f!=NULL)  
    fclose(f);
}
