/* user-upload.c
 *
 * Copyright (C) 2001 Michael McTernan,
 *   Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "user.h"
#include "timedread.h"

static void malformed(void);
static void pageend(void);
static int readLine(int sock,char **result);
static int openuserfile(const char *file);
static void gobble(void);

int main(int argc,char *argv[]) {
  char *boundary,*line,*filename;
  int   fd,tarball;
  
  /* Page title */
  printf("<h2>File Upload</h2>");

  /* Read first line, check it's a MIME boundary */
  if(readLine(0,&boundary)!=0 || strncmp("--",boundary,2)!=0) {
    free(boundary);
    gobble();
    malformed();
    return 0;
  }

  /* Cut off the "\r\n" */
  boundary[strlen(boundary)-2]='\0';
  for(fd=strlen(boundary)-1;fd>0;fd--)
    boundary[fd+2]=boundary[fd];

  boundary[0]='\r';
  boundary[1]='\n';

  /* Scan for filename while disposing MIME header */
  filename=NULL;

  while(readLine(0,&line)==0 && strcmp(line,"\r\n")!=0) {
    char *fstart,*fend;

    /* Look for filename if not yet found */
    if(filename==NULL) {

      fstart=strstr(line,"filename=\"");
      if(fstart!=NULL) {
        int t;

        fstart+=sizeof("filename=\"")-1; 
        fend=strstr(fstart,"\"");
   
        if(fend!=NULL)
          *fend='\0';
        else {
          free(boundary);
          free(line);
          malformed(); 
          gobble();
          return 0;
        }

        /* Some browsers send the whole path, so strip it off */
        t=strlen(fstart);
        while(t>0 && fstart[t-1]!='/' && fstart[t-1]!='\\')
          t--;
        fstart=&fstart[t];

        filename=strdup(fstart);
      }
    }

    free(line);

  } 

  /* Ensure we found the filename */
  if(filename==NULL) {
    free(boundary);
    gobble();
    malformed();
    return 0;
  }

  /* Attempt to open the file */
  fd=openuserfile(filename);

  if(fd<=0) {
    printf("<p>Could not open output file %s. The error was \"%m\"</p>\n",
           filename);
  } else {
    char *buffer;
    int   t,u,fsize=0;

    /* Allocate small buffer */
    buffer=malloc(strlen(boundary));

    /* Copy input out to file */
    t=0;
    do {
      u=timedread(0,&buffer[t++],sizeof(char));
      if(u==sizeof(char)) {
        
        if(buffer[t-1]!=boundary[t-1]) {
          fsize+=write(fd,buffer,t);
          t=0;
        }

      }
    } while(t<strlen(boundary) && u==sizeof(char));

    free(buffer);
    close(fd);
    
    /* Look for further section for tarball option */
    tarball=0;
    if(u==sizeof(char)) {
      
      while(readLine(0,&line)==0 && 
            strncmp(line,boundary,strlen(boundary))!=0) 
      {
        if(strncmp(line,"on",2)==0)
          tarball=1;
      }

    }

    /* Depending on whether we untar or not, display result */
    if(!tarball) {

      printf("<p>Written %d bytes to <a href=\"" USER_WWW_VROOT "%s\">"
             "%s</a> file.</p>",
             fsize,filename,filename);

    } else {
      char *cmd;

      printf("<p>Received %d bytes of tarball.</p>",fsize);
      
      printf("<p>Extracted files:<br><ul>\n");
      fflush(stdout);

      cmd=malloc(strlen(filename)+200);
      strcpy(cmd,"gunzip -c " USER_ROOT);
      strcat(cmd,filename);

      // Award: Stupid escaping
      strcat(cmd, " | tar -xvf - | sed \"s/^x \\(.*$\\)/<li><a href=\\\"" USER_WWW_VROOT_ESC "\\1\\\">\\1<\\/a>/\"");

      system(cmd);
      
      printf("</ul></p>");    

      free(cmd);
      unlink(filename);
    }
  } 

  /* Tidy off and return */
  gobble();
  free(filename);
  free(boundary);
  pageend();

  return 0;
}


/*openuserfile*/
/* Attempt to open a file in the web root so that we can 
 *  write the uploaded file to disk.
 */
static int openuserfile(const char *file) {
  char *filepath;
  int   fd;

  /* Guard against submitting file with .. or / in name */
  if(strstr(file,"..")!=NULL || strstr(file,"/")!=NULL)
    return -1;

  /* Concatentate with the web root */
  filepath=malloc(strlen(file)+sizeof(USER_ROOT)+1);
  strcpy(filepath,USER_ROOT);
  strcat(filepath,file);

  /* Open the file (Permissions 0644) */
  fd=open(filepath,O_CREAT|O_TRUNC|O_WRONLY,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH); 

  /* Tidy */
  free(filepath);

  return fd;

}

static void malformed() {

  printf("<p>Malformed request</p>\n");
  pageend();

}

static void pageend() {
  printf("<ul><li><a href=\"/user-upload.html\">File Upload</a></li>\n"
         "<br>"
         "<li><a href=\"/\">Main menu</a></li></ul>\n");

}

/*readline*/
/* Read a line of text from sock, allocating and placing resulting 
 *  string in *result.  Returns -1 on error, 0 on success.  
 *  Assumes lines are terminated with "\r\n" and < 1k.
 */
static int readLine(int sock,char **result) {
  char   buf[1024];
  int    t;
 
  t=0;
  do {
 
    if(timedread(sock,&buf[t],sizeof(char))<=0)
      t=sizeof(buf);  /* Force exit */
 
    t++;
  } while(t<sizeof(buf)-1 && ( t<2 ||  (buf[t-2]!='\r' && buf[t-1]!='\n') ) );
 
  if(t>=sizeof(buf)-1)
    return -1;
 
  /* Allocate the memory and return */
  buf[t]='\0';
  *result=strdup(buf);
  return 0;
 
}

/*gobble*/
/* Read everything on stdin and throw it away.
 */
static void gobble() {
  char buffer[1024];

  while(timedread(0,&buffer,1024)>0);

}
