/* ipchains-save.c
 *
 * Small program that dumps ipchains rules as ipchains commands
 *  
 * Copyright (C) 2001 Michael McTernan,
 *   Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <netdb.h>
#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "ipfwc_kernel_headers.h" 
#include "libipfwc.h"
#include "ipchains-save.h"

/* Prototypes */
static void dump_rules(FILE *out);
static void print_policy(FILE *out, ip_chainlabel ipcl);
static void print_addrmask(FILE *out, struct in_addr adr,struct in_addr mask,int inv);
static void print_protocol(FILE *out,int proto,int inv);
static void print_interface(FILE *out,ip_chainlabel inf,int inv,int wild); 
static void print_ports(FILE *out,int p1,int p2,int inv); 
static void print_flag(FILE *out,const char *flg,int cond,int inv);
static void print_flags(FILE *out,int flags,int invflags);
static char *getchainpolicy(const char *chain);

#define RULESFILE "/etc/rc.d/rc.firewall"
static const char *boundary="# ipchains entries go here";

void ipchains_save() {
  FILE *fi,*fo;
  int   p;
  char  c;

  /* Attempt to open the rule file */
  fi=fopen(RULESFILE,"r");
  if(fi==NULL) {
    printf("<p>Failed to open " RULESFILE ": %m</p>\n");
  } else {
    int t;

    /* Open temporary file to output */
    t=open("/tmp/rc.firewall",O_CREAT|O_WRONLY|O_TRUNC,S_IRWXU|S_IRGRP|S_IXGRP);
    
    fo=fdopen(t,"w");   

    /* Copy input file into output file until boundary or EOF */
    p=0;
    do {
      c=fgetc(fi);
      if(c!=EOF) {
       fputc(c,fo);
       if(c==boundary[p])
         p++;
       else
         p=0; 
      }
    } while(boundary[p]!='\0' && c!=EOF);     

    /* Done with input file */
    fclose(fi);

    /* Put the rules on */
    fprintf(fo,"\n"); 
    dump_rules(fo);
    fclose(fo);    
    unlink(RULESFILE);
    system("mv /tmp/rc.firewall " RULESFILE);
  }

  return; 
 
}
 


/*dump_rules*/
/* Dump the rules in chain 
 */
static void dump_rules(FILE *out) {
  struct ipfwc_fwrule *rules;
  char                *s;
  int                  nrules;
  int                  proto;
  int                  t;

  /* Get rules */
  rules=ipfwc_get_rules(&nrules,0);
  if(rules==NULL) {
    printf("Error fetching rules.");
    return;
  }

  /* First flush chains */
  fprintf(out,"ipchains -F %s\n",IP_FW_LABEL_INPUT);
  fprintf(out,"ipchains -F %s\n",IP_FW_LABEL_OUTPUT);
  fprintf(out,"ipchains -F %s\n",IP_FW_LABEL_FORWARD);


  /* Process rules */
  for(t=0;t<nrules;t++) {

    /* Print the ipchains command */ 
    fprintf(out,"ipchains -A %s ",rules[t].chain->label);
   
    proto=rules[t].ipfw.ipfw.fw_proto;
    print_protocol(out,proto,rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_PROTO);

    /* Source address/mask */
    if(rules[t].ipfw.ipfw.fw_smsk.s_addr!=0L) {
      fprintf(out,"-s ");    
      print_addrmask(out,
                     rules[t].ipfw.ipfw.fw_src,
                     rules[t].ipfw.ipfw.fw_smsk,
                     rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_SRCIP);
    }

    /* Destination address/mask */
    if(rules[t].ipfw.ipfw.fw_dmsk.s_addr!=0L) {
      fprintf(out,"-d ");
      print_addrmask(out,
                     rules[t].ipfw.ipfw.fw_dst,rules[t].ipfw.ipfw.fw_dmsk,
                     rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_DSTIP);
    }

    /* Only print ports if tcp or udp */
    if(proto==IPPROTO_TCP || proto==IPPROTO_UDP) { 

      /* Source port */
      if(rules[t].ipfw.ipfw.fw_spts[0]!=0 || 
         rules[t].ipfw.ipfw.fw_spts[1]!=65535) 
      {
        fprintf(out,"--sport ");    
        print_ports(out,
                    rules[t].ipfw.ipfw.fw_spts[0],
                    rules[t].ipfw.ipfw.fw_spts[1],
                    rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_SRCPT);
      }

      /* Destination port */
      if(rules[t].ipfw.ipfw.fw_dpts[0]!=0 ||
         rules[t].ipfw.ipfw.fw_dpts[1]!=65535)
      {
        fprintf(out,"--dport ");
        print_ports(out,
                    rules[t].ipfw.ipfw.fw_dpts[0],
                    rules[t].ipfw.ipfw.fw_dpts[1],
                    rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_DSTPT);
      }

    }
 
    /* Print the interface */
    print_interface(out,
                    rules[t].ipfw.ipfw.fw_vianame,
                    rules[t].ipfw.ipfw.fw_invflg & IP_FW_INV_VIA,
                    rules[t].ipfw.ipfw.fw_flg & IP_FW_F_WILDIF); 

    /* Add any additional flags */
    print_flags(out,rules[t].ipfw.ipfw.fw_flg,rules[t].ipfw.ipfw.fw_invflg);

    /* The policy for the rule */
    print_policy(out,rules[t].ipfw.label); 

    fprintf(out,"\n");
  }

  /* Add the default policies */
  fprintf(out,
          "ipchains -P %s %s\n",
          IP_FW_LABEL_INPUT,
          s=getchainpolicy(IP_FW_LABEL_INPUT));
  free(s);
  fprintf(out,
          "ipchains -P %s %s\n",
          IP_FW_LABEL_OUTPUT,
          s=getchainpolicy(IP_FW_LABEL_OUTPUT));
  free(s);
  fprintf(out,
          "ipchains -P %s %s\n",
          IP_FW_LABEL_FORWARD,
          s=getchainpolicy(IP_FW_LABEL_FORWARD));
  free(s);
 
}


/*getchainpolicy*/
static char *getchainpolicy(const char *chain) {
  struct ipfwc_fwchain *chains;
  int                   nchains;
  int                   t;

  /* Get all the chains */
  chains=ipfwc_get_chainnames(&nchains);

  for(t=0;t<nchains;t++) {
    
    if(strcmp(chains[t].label,chain)==0) 
      return strdup(chains[t].policy);
   
  }
  
  return NULL;
} 


/*print_protocol*/
/* Print the name of a protocol 
 */
static void print_protocol(FILE *out,int proto,int inv) {
  struct protoent *pe;

  /* Special case for 'any' */
  if(proto==0) 
    return;

  fprintf(out,"-p ");
 
  if(inv)
    fprintf(out,"! ");

  /* Try system approach */
  pe=getprotobynumber(proto); 
  if(pe!=NULL) {
    fprintf(out,"%s ",pe->p_name);
    // TODO: free(pe); ????
  } else {

    /* Hard wire common ones incase /etc/protocols is broken */
    switch(proto) {
    case IPPROTO_TCP:
      fprintf(out,"tcp ");
      break;
    case IPPROTO_UDP:
      fprintf(out,"udp ");
      break;
    case IPPROTO_ICMP:
      fprintf(out,"icmp ");
      break;
    default:
      fprintf(out,"%u ",proto);
    }

  }


}


/*print_addrmask*/
/* Print the address/mask in a pretty way
 */
static void print_addrmask(FILE *out,
                          struct in_addr adr,
                          struct in_addr mask,
                          int inv) 
{
  char *a=inet_ntoa(adr);

  if(inv)
    fprintf(out,"! ");

  fprintf(out,"%s",a);

  /* Add mask if valid */
  if(ntohl(mask.s_addr)!=0xFFFFFFFFL)
    fprintf(out,"/%s",inet_ntoa(mask));

  fprintf(out," ");

}


/*print_policy*/
/* Print a ipchain policy or rule destination 
 */
static void print_policy(FILE *out,ip_chainlabel ipcl) {

  fprintf(out,"-j %s",ipcl);

}


/*print_interface*/
/* Print an interface name.
 */
static void print_interface(FILE *out,ip_chainlabel inf,int inv,int wild) {

  if(*inf=='\0')
    return;

  if(inv)
    fprintf(out,"-i ! %s",inf);
  else
    fprintf(out,"-i %s",inf);

  if(wild)
    fprintf(out,"+");
 
  fprintf(out," ");
}

static void print_flags(FILE *out,int flags,int invflags) {

  print_flag(out,"-l ",flags & IP_FW_F_PRN,0);
  print_flag(out,"-y ",flags & IP_FW_F_TCPSYN,invflags & IP_FW_INV_SYN);
  print_flag(out,"-f ",flags & IP_FW_F_FRAG,invflags & IP_FW_INV_FRAG);

}

static void print_flag(FILE *out,const char *flg,int cond,int inv) {
  if(cond) {
    if(inv)
      fprintf(out,"! ");
    fprintf(out,flg);
  }
}


/*print_ports*/
/* Print the port range 
 */
static void print_ports(FILE *out,int p1,int p2,int inv) {

  if(inv)
    fprintf(out,"! ");

  if(p1==p2)
    fprintf(out,"%d ",p1);
  else
    fprintf(out,"%d:%d ",p1,p2);  

}

