/* request.c
 *
 * Copyright (C) 2001 Michael McTernan,
 *   Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include "request.h"
#include "timedread.h"

struct request {
  char *url;
  char *clear_url;
  char *method;
  char *params;
};

static int readWord(int sock,char **result);
static int skipHeader(int sock);

/*REQ_New*/
/* Create new request */
int REQ_New(int sock,request_t *req) {

  /* Allocate for the ADT structure */
  *req=malloc(sizeof(struct request));
  if(*req==NULL)
    return -1;

  /* Read the request method */
  if(readWord(sock,&(*req)->method)!=0) {
    free(*req);
    return -1;
  }

  /* Read the whole URL */
  if(readWord(sock,&(*req)->url)!=0) {
    free((*req)->method);
    free(*req);
    return -1;
  }

  /* Ignore rest of HTTP header */
  skipHeader(sock);

  /* Look for parameter string */
  (*req)->params=strstr((*req)->url,"?");
  if((*req)->params!=NULL) {
    *(*req)->params='\0';
    (*req)->params++;
  }

  /* Calculate this when needed */
  (*req)->clear_url=NULL;

  return 0;
  
}


/*REQ_GetMethod*/
const char *REQ_GetMethod(request_t req) {
  return req->method;
}


/*REQ_GetURL*/
const char *REQ_GetURL(request_t req) {
  return req->url;
}


/*REQ_GetDemangledURL*/
const char *REQ_GetDemangledURL(request_t req) {
  int t,u;

  /* Check if we have done this before */    
  if(req->clear_url!=NULL)
    return req->clear_url;

  /* First duplicate the url */
  req->clear_url=strdup(req->url);

  /* Copy char by char, including null */   
  for(u=t=0;t<strlen(req->url)+1;t++,u++) {
					
    req->clear_url[u]=req->url[t];
    
		/* Check for escape code */
    if(req->url[t]=='%') {
	    char buf[3];
			int  value;
			
			/* Get the 2 digit hex code, parse and write */
		  strncpy(buf,&req->url[t+1],2);
      sscanf(buf,"%x",&value); 
			req->clear_url[u]=value;
			
		  /* Skip the next 2 digits */
			t+=2;
    }
  }
  
  return req->clear_url;

}


/*REQ_GetQuery*/
const char *REQ_GetQuery(request_t req) {
  return req->params;
}

/*REQ_Destroy*/
void REQ_Destroy(request_t *req) {
  free((*req)->method);
  free((*req)->url);
  free((*req)->clear_url);
  /* Do not free params */
  free((*req));
  *req=NULL;
}


/*readWord*/
/* Read from sock until white space is encountered.
 * Return 0 for success, otherwise an error occured.
 */
static int readWord(int sock,char **result) {
  char   buf[1024];
  int    t;
 
  t=0;
  do {
 
    if(timedread(sock,&buf[t],sizeof(char))<=0)
      t=sizeof(buf);  /* Force exit */
 
    t++;
  } while(t<sizeof(buf) && !isspace(buf[t-1]));
 
  if(t>=sizeof(buf))
    return -1;

  /* Allocate the memory and return */
  buf[t-1]='\0';
  *result=strdup(buf);
  return 0;

}

/*skipHeader*/
/* Skip HTTP headers.
 */
static int skipHeader(int sock) {
  char end[]="\r\n\r\n";
  int  p=0;

  do {
    char c;

    if(timedread(sock,&c,sizeof(char))<=0)
      return -1;

    p= (c==end[p]) ? p+1 : 0; 
    
  } while(p<4);

  return 0;
}

