/* webctl.c 
 *
 * HTTP daemon for controlling Coyote Linux distros. 
 * Copyright (C) 2001 Michael McTernan,
 *   Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h> 
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include "utils.h"
#include "request.h"
#include "dynamictable.h"
#include "messages.h"

#ifdef USER_SERVER
#include "user.h"
#endif

static void sendData(int sock,const char *s,int len);
static void checkTable(int sock,request_t req);
static int handleConnection(int sock);
static int makeSocket();


static int psock;

int main(int argV,char *argC[]) {
 
  /* Create the socket */
  psock=makeSocket();

  /* Ignore children */
  signal(SIGCHLD,SIG_IGN);

  do {
    struct sockaddr_in caddr;
    socklen_t          caddrlen;
    int csock;

    /* Accept connection */
    caddrlen=sizeof(struct sockaddr_in);
    csock=accept(psock,(struct sockaddr *)&caddr,&caddrlen);
    warning(csock!=-1);

    if(csock!=-1) {
      handleConnection(csock);
      close(csock);
    }  

  } while(1);

}


/*handleConnection*/
/* Handle a connection on some previously accepted socket.
 */
static int handleConnection(int sock) {
  request_t request;

  /* Parse header */
  if(REQ_New(sock,&request)!=0)
    return -1;

  /* Check the request method */
  if(strcasecmp("HEAD",REQ_GetMethod(request))==0) {
    MSG_Header(sock,501,"Not Implemented");
    MSG_Body(sock,"Not Implemented");
  } else {

    #ifdef USER_SERVER
      if(strncmp(USER_WWW_VROOT,REQ_GetURL(request),strlen(USER_WWW_VROOT))==0)
        checkUser(sock,request);
      else
    #endif

    checkTable(sock,request);  

  }

  /* Free memory */
  REQ_Destroy(&request);
  return 0;

}


/*checkTable*/
/*
 */
static void checkTable(int sock,request_t req) {
  char  *requri,*qmark;
  const  struct table_entry *tentry;

  /* Get request URI */
  requri=strdup(REQ_GetDemangledURL(req));

  /* Cut off options */
  qmark=strstr(requri,"?");
  if(qmark!=NULL)
    *qmark='\0';

  tentry=content_table;

  /* Scan table for a matching entry */
  while(tentry->url!=NULL && strcmp(tentry->url,requri)!=0)
    tentry++;
 
	free(requri);
 	
  /* Return not found if it isn't in the table */
  if(tentry->url==NULL) {
    MSG_Error(sock,404,"Not Found");
    return;
  }

  switch(tentry->type) {
  case page_action_code:
    MSG_Header(sock,200,"Ok");
    MSG_PageHeader(sock);
    sendData(sock,tentry->u.page.data,tentry->u.page.length);
    MSG_PageFooter(sock);     
    return;
  case exec_action_code:
  case cgi_action_code:
    
    MSG_Header(sock,200,"Ok"); 
    MSG_PageHeader(sock);
 
    /* Set the QUERY_STRING env variable */
    if(REQ_GetQuery(req)!=NULL) {
      setenv("QUERY_STRING",REQ_GetQuery(req),1);
    } else
      setenv("QUERY_STRING","",1);    
 
    /* Execute as a new process */
  #ifdef NO_FORK
    if(1) {
  #else
    if(fork()==0) {
  #endif

      /* Set the stdin,stdout etc to socket */
      dup2(sock,0);
      dup2(sock,1);      
      dup2(sock,2);
 
      /* Close other fds */
      close(sock);
    #ifndef NO_FORK
      close(psock);  /* Main listening socket: cgi can't steal connections */
    #endif

			/* Execute command */
      if(tentry->type==cgi_action_code) {
        system(tentry->u.cgi.exec_string);
      } else {
        tentry->u.exec.func(0,NULL);
      } 

			/* Ensure any printf's are flushed */
      fflush(stdout);
      fflush(stderr);

      MSG_PageFooter(1);   
      exit(0);

    }
  
  }

}


/*sendData*/
/* Send a whole buffer of data */
static void sendData(int sock,const char *s,int len) {
  int t=0;

  while(t<len) {
    int c=write(sock,&s[t],len-t);
    if(c<=0)
      return;
    else
      t+=c;
  }

}


/*makeSocket*/
/* Make a socket and bind to some local address.
 */
static int makeSocket() {
  struct sockaddr_in addr;
  int tru=1;  
  int s;

  /* Create the socket */
  critical((s=socket(PF_INET,SOCK_STREAM,0))!=-1);

  /* Create the address */
  addr.sin_family=AF_INET;
  addr.sin_port=htons(888);
  addr.sin_addr.s_addr=INADDR_ANY;  /* Anyone can connect */

  /* Set to reuse address */
  warning(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&tru,sizeof(int))==0);
 
  /* Give it an address */
  critical(bind(s,(struct sockaddr *)&addr,sizeof(struct sockaddr_in))==0);

  /* Set the listen queue */
  listen(s,40);
 
  return s;

}

